/*
    Biribot.ino
    Release 1.0

  Hardware needed:
  
    - Arduino nano
    - 2 TowerPro s 9g servos modified for continuous rotation
    - HCSR04 ultrasonic sensor
    
  ToDo;

    Add magnetic compass sensor
    Add IR sensors  

  Avalaible Actions to be programmed in loop()
      
  Move( FORWARD , V_ACT );
  Move( REVERSE , V_ACT );

  Rotate( RIGHT , V_ACT );
  Rotate( LEFT , V_ACT );

  Turn( LEFT_FW , V_ACT );
  Turn( LEFT_RV , V_ACT );
  Turn( RIGHT_FW , V_ACT );
  Turn( RIGHT_RV , V_ACT );

  LM_Test();
  RM_Test();
  Drivetrain_test();

  PAUSE_1;
  PAUSE_2;
  PAUSE_5;

  FOREVER;

*/




/*
  The following line must be set to
  #define DEBUG or #define RELEASE or #define TEST - See loop() -
  
  DEBUG uses the Serial Monitor and Imprime() function
  TEST runs Debug_Test() funcion and stops
  RELEASE executes the avoiding obstacles program
*/


#include "Biribot.h"

#define RELEASE

//   Servos ( drive train servos, left-motor & right-motor )
#include <Servo.h> 
Servo lm,rm;





/*

    setup --> only runs once b4 loop()

    Basically, the structure of a program using sketches is  ( Watch main.cpp in cores directory )
    
    void main( void)
    {
      init();
      setup();
      for(;;)
        loop()
    }
*/
void setup() 
{ 
  // Ultrasonic HCSR04
  pinMode(HCSR04_TRIG_PIN, OUTPUT);
  pinMode(HCSR04_ECHO_PIN, INPUT);

  #if defined RELEASE
    Engine_start();

    PAUSE_5;          // Nice to wait for a while b4 moving ...
  #endif
  #if defined DEBUG
    Serial.begin(9600);
  #endif 
} 



/*

    main loop

*/
void loop() 
{ 
  static float ant , pos;

  //  Drivetrain Test. Change the first #define in the code to #define TEST
  #if defined TEST
    Drivetrain_test();
    FOREVER;                  //   Stop here
  #endif

  //  Debug version. Change the first #define in the code to #define DEBUG
  //  Run it to test and tune the HCSR04 sensor
  //  Watch notes in Read_distance function
  #if defined DEBUG
    FOREVER
    {
       pos = Read_distance();
       Imprime( pos);
       PAUSE_1;
    }
  #endif
  
  //  Release version. Change the first #define in the code to #define RELEASE
  #if defined RELEASE

    //  0) Go ahead until we get to DIST_HORIZONTE of an object

    Move( FORWARD , V_ACT );
    do 
    {
      pos = Read_distance();
      // Feedback code for servos here
    } while( pos > DIST_HORIZONTE );
    Neutral();
    PAUSE_1;

    /*
        1)  We try to get the normal direction to the object  ( we may have a great incident angle and this is be a problem ),
        since a great incident angle means sensor out of range and Read_distance() returns (ie) 123 cm
        To avoid this we should make a sweep left and right constantly 
        In other words, the sensor returns the same value if Biri is at 123 cm of a wall, or nearly parallel to the wall at 1 cm  
        before to crash aganinst the wall 
    */

    Moveto_normal2(0.5,80);   //These pars work fine    Take care if you change them
  
    Neutral();
    PAUSE_1;

    //  2) Go ahead until we get to DIST_MIN_MAX of an object now in the normal direction to it

    pos=Read_distance();
    if( pos <  DIST_MIN_MAX )
    {
      Move( REVERSE , V_ACT );
      do 
      {
        pos = Read_distance();
        delay(50);
          // Feedback code for servos here
      } while( pos < DIST_MIN_MAX );
    }
    else
    {
      Move( FORWARD , V_ACT );
      do 
      {
        pos = Read_distance();
        delay(50);
          // Feedback code for servos here
      }while( pos > DIST_MIN_MAX );
    }
    Neutral();
    PAUSE_1;
/*
    // 3)  Re-find the normal //We can avoid this step
    Moveto_normal2(0.5, 50);
    Neutral();
    PAUSE_1;
*/        
                // Avoiding strategy code starts HERE points 1) to 4)
    
    //  A) We go to DIST_MIN_MIN slowly, in a different and funny do-while loop     ... 

    for( Move( FORWARD , V_MIN),pos=Read_distance(), ant=pos+0.5 ; pos > DIST_MIN_MIN ; pos= Read_distance() )
    {
      // Here we fill the vector of IR sensor measures (In a future post with IR )
      //Read_distance();  
      
      // Update  ant ( previous minimun distance )
      if( pos < ant )
        ant=pos;
    }
    Neutral();
    PAUSE_1;
    
    // B) ... Move back to DIST_HORIZONTE
    Move( REVERSE , V_ACT );
    do 
    {
         // Feedback code for servos here
    } while(  (pos = Read_distance() ) <= DIST_HORIZONTE );
    Neutral();
    PAUSE_1;
        
    // C) ... Turn right and ...
    Rotate( RIGHT , V_ACT );
    PAUSE_1;
    
    // ... D) and stop for a second and start again the loop
    Neutral();
    PAUSE_1;
        
  #endif  

} 
/*
    
      End of loop()


   __________________________________________________________________________________________________________________________
*/



/*
    Moveto_normal()

      Finds the normal direction to an obstacle ( ie a wall) and aligns Biri this direction
      A very tricky function but its difficult to do anything else with only with a HCSR04
      If we had connected our HCSRO4 sensor to a servo, this would have been easy & simple ,but this brigde must be crossed 
      in another bot
  
      The normal is the minimum distance between 2 points and we will try to get the minimun 

  Notes: It was a bit boooring to adjust and tune the code ...
         Hope nobody will have to uncomment the Serial.lines ...  this will mean you are in trouble .... or researching hahahaha
*/
void Moveto_normal2( float precision , int time )
{
  float pos, min_distance;           
  int sidetoturn , i;  

  //  Set up the minimum distance to Biri's actual position
  min_distance=Read_distance();
  
  // Give one step LEFT and revalue distance comparing mindistance and pos
  Rotate( LEFT , V_ROT );    
  delay( time );
  Neutral();
  pos = Read_distance();

  if( Are_near(pos , min_distance , precision) )    //  We are near normal direction to obstacle so we can finish 
    return;
  
  if( pos < min_distance ) 
    sidetoturn = LEFT;      // We have a shorter distance to the LEFT, so we Keep on this direction
  else
    sidetoturn = RIGHT;      // We are going farther, so the correct turn is to the RIGHT
    
  for( min_distance=1000 , i=0; ; i++)
  {
    switch( sidetoturn )
    {
      case LEFT : Rotate( LEFT , V_ROT ); break;    // One more step LEFT until we situate Biri farther than he is now 
      case RIGHT: Rotate( RIGHT , V_ROT ); break;   // One more step RIGHT until we situate Biri farther than he is now 
      default: return; 
    }
    delay(time);
    Neutral();
    pos = Read_distance();

    if( Are_near(pos , min_distance , precision) )    //  We are near normal direction to obstacle so we can finish 
      break;
    
    /*
        Update min_distance. Mathly speaking we are finding the minimun distance in a decreasing function ( Function distance )
        ... and the first higher value we find, the function changes its sign, so we go.
    */
     min_distance = pos;

  }
}




/*
    Are_near()
    
      Helping hand to Moveto_normal2()

    Notes: SI & NO are in Biribot.h. Take a glance at two funny and advanced #include
*/
int Are_near( float a, float b , float error )
{
  float aux = abs(a-b);

//      Serial.print("Min_distance...");Serial.print(min_distance);Serial.print("\tPos ...");Serial.print(pos);
//      Serial.print("\tAbs...");Serial.println(abs( pos - min_distance) );

  if( aux < error )
    return SI;
  else
    return NO;
  }







/*
    Read_distance( void )

    Reads the distance with a 2 decimals
    
    Notes:
      Change the const HCSR04_K in Biribot.h and tune it first with #define DEBUG before running the program
    
      The standard value is 58.0 but my sensor works fine with 80.0
      You can use any kind of library such as "Newping.h" but change it to give 2 decimals
      If not, Movetonormal2() will fail clairly
*/

float Read_distance(void)
{
  unsigned long pulso, aux;

  delay(30);
  // Trigger pulse in HCSR04 sensor. We need only a 5/10 microsg pulse
  digitalWrite(HCSR04_TRIG_PIN, LOW);
  digitalWrite(HCSR04_TRIG_PIN, HIGH);
  delayMicroseconds(HCSR04_PULSEWIDTH);
  digitalWrite(HCSR04_TRIG_PIN, LOW);
    
  //Read the Echo if exists
  pulso = pulseIn(HCSR04_ECHO_PIN, HIGH);  //pulso = pulseIn(HCSR04_ECHO_PIN, HIGH,1500); 1500 shortens the range to 4 cm - 20 cm

  #if defined DEBUG
    Serial.print( "   Pulso ... ");
    Serial.print( pulso  );
  #endif
    
  //This gives 2 decimals as a float  
  aux= (int ) 100*(0.5+pulso/HCSR04_K ) ;
  return (float) (aux/100.0);
}




#if defined DEBUG

/*
    Imprime(( float distancia )
*/
void Imprime ( float distancia )
{
  
  Serial.print( "   Distancia ... ");
  Serial.println( distancia , NUM_DEC );
}

#endif




/*
    Engine_start()
*/
void Engine_start( void )
{
  lm.attach(LM_PIN);    
  rm.attach(RM_PIN);  
  Neutral();
}



/*
    Engine_stop()
*/
void Engine_stop( void )
{
  Neutral();
  lm.detach();    
  rm.detach();  
}



/*
    Move()
    
    Move Biribot moving both motors the same direction
    
      pars:
        where: { FORWARD , REVERSE }
        speed : { 0 , V_MAX  }

    Note: We can try Biribot to move straight using different speeds for lm and rm, ie if Biri goes right ... 

      lm_move( FORWARD , speed-5);    we reduce left motor speed, so the wheel will turn slower 
      rm_move( FORWARD , speed);  

    This is somekind of feedback and may be ok if we dont have a compass
*/
void Move( int where , int speed )
{
  switch( where )
  {
    case FORWARD:
      lm_move( FORWARD , speed);  //Retoque
      rm_move( FORWARD , speed);
      break;
    case REVERSE:
      lm_move( REVERSE , speed);  //Retoque
      rm_move( REVERSE , speed);
      break;
  }
}



/*
    Turn()
    
    Turn Biribot moving only one motor FORWARD/REVERSE
    
      pars:
        where: { LEFT_FW , LEFT_RV , RIGHT_FW , RIGHT_RV }     ( FW: FORWARD , RV: REVERSE )
        speed : { 0 , V_MAX  }
*/
void Turn( int where , int speed )
{
  switch( where )
  {
    case LEFT_FW:
      lm_move( STOP , speed);
      rm_move( FORWARD , speed);
      break;
    case LEFT_RV:
      lm_move( STOP , speed);
      rm_move( REVERSE , speed);
      break;
    case RIGHT_FW:
      lm_move( FORWARD , speed);
      rm_move( STOP , speed);
      break;
    case RIGHT_RV:
      lm_move( REVERSE , speed);
      rm_move( STOP , speed);
      break;
  }
}



/*
    Rotate()
    
    Rotate Biribot moving both motor opposite directions ( ie lm FORWARD , rm REVERSE )
    
      pars:
        where: { LEFT , RIGHT }
        speed : { 0 , V_MAX  }
*/
void Rotate( int where , int speed )
{
  switch( where )
  {
    case LEFT:
      lm_move( REVERSE , speed);
      rm_move( FORWARD , speed);
      break;
    case RIGHT:
      lm_move( FORWARD , speed);
      rm_move( REVERSE , speed);
      break;
  }
  delay( DT_ROTATE );
}



/*
    Neutral()

    Stops motors moving each to center position    
*/
void Neutral(void)
{
  lm.write(LM_CENTER);
  rm.write(RM_CENTER);
  delay( DT_NEUTRAL );
}




/*******************************************************************************************************************/





/*
    lm_move()
    
    Move Left Motor 
    
      pars:
        where: { FORWARD , REVERSE , STOP }
        speed : { 0 , V_MAX  }
*/
void lm_move( int where , int speed )
{
  if( (speed > V_MAX) || (speed < V_MIN) )
    return;

  switch( where )
  {
    case FORWARD:
      lm.write(LM_CENTER+speed); 
      break;
    case REVERSE:
      lm.write(LM_CENTER-speed); 
      break;
    case STOP:
    default:
      lm.write(LM_CENTER);
  }
  delay(D_MOV);
}



/*
    rm_move()

    Move Right Motor 
    
      pars:
        where: { FORWARD , REVERSE , STOP }
        speed : { 0 , V_MAX  }
*/
void rm_move( int where , int speed )
{
  if( (speed > V_MAX) || (speed < V_MIN) )
    return;

  switch( where )
  {
    case FORWARD:
      rm.write(RM_CENTER-speed);
      break;
    case REVERSE:
      rm.write(RM_CENTER+speed);
      break;
    case STOP:
    default:
      rm.write(RM_CENTER);
  }
  delay(D_MOV);
}



/*
    Left Motor Test
*/
void LM_Test( void )
{
  lm_move( FORWARD , V_TEST );
  PAUSE_2;
  lm_move( STOP , V_TEST );
  PAUSE_2;
  lm_move( REVERSE , V_TEST );
  PAUSE_2;
  lm_move( STOP , V_TEST );
  PAUSE_2;
}



/*
    Right Motor Test
*/
void RM_Test( void )
{
  rm_move( FORWARD , V_TEST );
  PAUSE_2;
  rm_move( STOP , V_TEST );
  PAUSE_2;
  rm_move( REVERSE , V_TEST );
  PAUSE_2;
  rm_move( STOP , V_TEST );
  PAUSE_2;
}



/*
    Right Motor Test
*/
void Drivetrain_test( void )
{
  Move( FORWARD , V_ACT );
  PAUSE_5;
  Neutral();
  PAUSE_2;
  Move( REVERSE , V_ACT );
  PAUSE_5;
  Neutral();
  PAUSE_2;

  Rotate( RIGHT , V_ACT );
  PAUSE_2;
  Neutral();
  PAUSE_2;

  Rotate( LEFT , V_ACT );
  PAUSE_2;
  Neutral();
  PAUSE_2;

  Turn( LEFT_FW , V_ACT );
  PAUSE_2;
  Neutral();
  PAUSE_2;

  Turn( LEFT_RV , V_ACT );
  PAUSE_2;  
  Neutral();
  PAUSE_2;
  
  Turn( RIGHT_FW , V_ACT );
  PAUSE_2;  
  Neutral();
  PAUSE_2;
    
  Turn( RIGHT_RV , V_ACT );
  PAUSE_2;      
  Neutral();
  PAUSE_2;
}      


